/*
To test the method for calling CUDA C subroutine from FORTRAN main program
Caculating exponential function
Author Yu Shi
Date: 05/21/2010
 */

/* Includes, system */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>


__global__ void CUDAexp(const double*, double*, const int );
extern "C"
int gexp_(double*,double*, int*);
extern "C"
int initmap_();

int gexp_(double* a,double* d, int* N1)
{    
	double *b,*c;
        clock_t start1,end1;
        int N=N1[0];
	cudaError_t statusb,statusc;

        cudaEvent_t start, stop;
        cudaEventCreate(&start);
        cudaEventCreate(&stop);
        cudaEventRecord(start,0);
        start1 = clock();

	statusb=cudaHostGetDevicePointer((void **)&b, (void *)a, 0);
        statusc=cudaHostGetDevicePointer((void **)&c, (void *)d, 0);

	if (statusb != 0 || statusc !=0) {
		printf("Error when locating memory to arrays on device!\n");
                printf("%d,%d\n",statusb,statusc);
		return EXIT_FAILURE;	
	}

        
	int MM = N%81920;
        int M = (N-MM)/81920 ;

	for (int i=0; i<M+1; i++){
		CUDAexp<<<160,512>>>(b+i*81920,c+i*81920, N);
	}	
	cudaEventRecord(stop,0);
	cudaEventSynchronize(stop);

	cudaFree(b); 
	cudaFree(c); 

	float elapsedTime;
        cudaEventElapsedTime(&elapsedTime,start,stop);
        end1 = clock();
        printf("CUDA Operation:%e\n",elapsedTime/1e3);
	printf("CUDA Operation:%e\n",((double)(end1-start1))/CLOCKS_PER_SEC);

	return EXIT_SUCCESS;
}

int initmap_()
{
        cudaDeviceProp prop;
        cudaGetDeviceProperties(&prop, 0); 
        if (!prop.canMapHostMemory) exit(0);
        cudaSetDeviceFlags(cudaDeviceMapHost);
        return 1;
}


// Device code
__global__ void CUDAexp(const double* b, double* c, const int N) {
        int bx = blockIdx.x;
        int tx = threadIdx.x;
        int index = 512*bx+tx;
        if(index <= N)c[index] = exp(b[index]);
}

